# distutils: language = c++
"""
This cython module wraps the struct templates exposed from C++ by data_structures_cython.pxd into extension types.
However an extension type obviously cannot be templated (at compile time, all possible variations of a template
must be known). There is no really elegant way of doing it in cython as of v3.0a6. So we will use the [Explicit
Run-Time Dispatch approach](https://martinralbrecht.wordpress.com/2017/07/23/adventures-in-cython-templating/).

In short, an enum LocType (resides in data_structures.pxd) contains all template types Loc we are likely to need.
The extension dtype for C++ object e.g. CRequest[Loc] then will contain *an union* holding one of the variants (e.g.
CRequest[int], Request[tuple(double, double)]) etc. Then each member function will check the Loc type, and dispatch the
method call to the appropriate object inside that union.
"""

from numpy import inf

from cython.operator cimport dereference
from libcpp.vector cimport vector
from libcpp.memory cimport shared_ptr, make_shared
from libcpp.memory cimport dynamic_pointer_cast
from cython.operator cimport typeid


from ridepy.data_structures_cython.cdata_structures cimport (
    Request as CRequest,
    TransportationRequest as CTransportationRequest,
    InternalRequest as CInternalRequest,
    Stop as CStop,
    R2loc,
)


import logging
logger = logging.getLogger(__name__)

cdef class Request:
    """
    The base class, never to be instantiated.
    """
    def asdict(self):
        return dict(
            request_id=self.request_id,
            creation_timestamp=self.creation_timestamp,
        )

    def __repr__(self):
        return f"{self.__class__.__name__}(" + ", ".join(f"{k}={v}" for k, v in self.asdict().items()) + ")"

    property request_id:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self._ureq._req_r2loc).request_id
            elif self.loc_type == LocType.INT:
                return dereference(self._ureq._req_int).request_id
            else:
                raise ValueError("This line should never have been reached")
        def __set__(self, value):
            if self.loc_type == LocType.R2LOC:
                dereference(self._ureq._req_r2loc).request_id = value
            elif self.loc_type == LocType.INT:
                dereference(self._ureq._req_int).request_id = value
            else:
                raise ValueError("This line should never have been reached")

    property creation_timestamp:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self._ureq._req_r2loc).creation_timestamp
            elif self.loc_type == LocType.INT:
                return dereference(self._ureq._req_int).creation_timestamp
            else:
                raise ValueError("This line should never have been reached")
        def __set__(self, value):
            if self.loc_type == LocType.R2LOC:
                dereference(self._ureq._req_r2loc).creation_timestamp = value
            elif self.loc_type == LocType.INT:
                dereference(self._ureq._req_int).creation_timestamp = value
            else:
                raise ValueError("This line should never have been reached")


    def __dealloc__(self):
        """
        Using smart pointers. Do not need to delete the base pointer.
        """
        ...


cdef class TransportationRequest(Request):
    """
    Cython equivalent of pure python dataclass :class:`.data_structures.TransportationRequest`. See its docstring for
    details of its attributes.

    .. admonition:: Implementation details for developers

       Wraps the C++ templated struct `data_structures_cython.cdata_structures.pxd::TransportationRequest[Loc]` into a
       cython extension type. The template parameter `Loc` defines the type of location objects (`Tuple[float, float]`
       for simulations in :math:`\mathbb{R}^2`), `int` for graphs.

       However a cython extension type obviously cannot be templated (at compile time, all possible variations of a template
       must be known). So we will use the Explicit Run-Time Dispatch approach, see this module's docstring or :ref:`desc_runtime_dispatch`
       for details.

       Holds a `shared_ptr` to C++ templated class `TransportationRequest<Loc>`. This class can be instantiated from an
       existing `shared_ptr` (only from the cython side and not from the calling python code). See the staticmethods
       `cdef TransportationRequest from_c_r2loc` and `cdef TransportationRequest from_c_int` for details.
    """
    def __init__(
            self,
            int request_id,
            double creation_timestamp,
            origin,
            destination,
            pickup_timewindow_min=0,
            pickup_timewindow_max=inf,
            delivery_timewindow_min=0,
            delivery_timewindow_max=inf,
    ):
        """
        Note
        ----
        We will try to infer `Loc` from what `origin` contains. The inference can be easily
        fooled by passing e.g. `origin = 2, destination = (0.3, 1.5)`.
        """
        if hasattr(origin, '__len__') and len(origin) == 2:
            # TODO: this inferring of LocType is kludgy. We should have it as an argument of __init__
            # let's assume both origin and destination are Tuple[double, double]
            self.loc_type = LocType.R2LOC
            self._utranspreq._req_r2loc = shared_ptr[CTransportationRequest[R2loc]](new CTransportationRequest[R2loc](
                request_id, creation_timestamp, origin, destination,
                pickup_timewindow_min, pickup_timewindow_max,
                delivery_timewindow_min, delivery_timewindow_max
            ))
            self._ureq._req_r2loc = dynamic_pointer_cast[CRequest[R2loc], CTransportationRequest[R2loc]](self._utranspreq._req_r2loc)
        elif isinstance(origin, int):
            # let's assume both origin and destination are int
            self.loc_type = LocType.INT
            self._utranspreq._req_int = shared_ptr[CTransportationRequest[int]](new CTransportationRequest[int](
                request_id, creation_timestamp, origin, destination,
                pickup_timewindow_min, pickup_timewindow_max,
                delivery_timewindow_min, delivery_timewindow_max
            ))
            self._ureq._req_int = dynamic_pointer_cast[CRequest[int], CTransportationRequest[int]](self._utranspreq._req_int)
        else:
            raise TypeError(f"Cannot handle origin of type {type(origin)}")

    def asdict(self):
        """
        Returns a dict representation of itself.
        """
        return dict(
            request_id=self.request_id,
            creation_timestamp=self.creation_timestamp,
            origin=self.origin,
            destination=self.destination,
            pickup_timewindow_min=self.pickup_timewindow_min,
            pickup_timewindow_max=self.pickup_timewindow_max,
            delivery_timewindow_min=self.delivery_timewindow_min,
            delivery_timewindow_max=self.delivery_timewindow_max
        )

    def __eq__(self, other: TransportationRequest):
        if not isinstance(other, TransportationRequest):
            return False
        return self.request_id == other.request_id \
            and self.loc_type == other.loc_type \
            and self.creation_timestamp == other.creation_timestamp \
            and self.origin == other.origin \
            and self.destination == other.destination \
            and self.pickup_timewindow_min == other.pickup_timewindow_min \
            and self.pickup_timewindow_max == other.pickup_timewindow_max \
            and self.delivery_timewindow_min == other.delivery_timewindow_min \
            and self.delivery_timewindow_max == other.delivery_timewindow_max

    def __repr__(self):
        return f"{self.__class__.__name__}(" + ", ".join(f"{k}={v}" for k, v in self.asdict().items()) + ")"


    property origin:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self._utranspreq._req_r2loc).origin
            elif self.loc_type == LocType.INT:
                return dereference(self._utranspreq._req_int).origin
            else:
                raise ValueError("This line should never have been reached")
        def __set__(self, value):
            if self.loc_type == LocType.R2LOC:
                dereference(self._utranspreq._req_r2loc).origin = value
            elif self.loc_type == LocType.INT:
                dereference(self._utranspreq._req_int).origin = value
            else:
                raise ValueError("This line should never have been reached")


    property destination:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self._utranspreq._req_r2loc).destination
            elif self.loc_type == LocType.INT:
                return dereference(self._utranspreq._req_int).destination
            else:
                raise ValueError("This line should never have been reached")
        def __set__(self, value):
            if self.loc_type == LocType.R2LOC:
                dereference(self._utranspreq._req_r2loc).destination = value
            elif self.loc_type == LocType.INT:
                dereference(self._utranspreq._req_int).destination = value
            else:
                raise ValueError("This line should never have been reached")



    property pickup_timewindow_min:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self._utranspreq._req_r2loc).pickup_timewindow_min
            elif self.loc_type == LocType.INT:
                return dereference(self._utranspreq._req_int).pickup_timewindow_min
            else:
                raise ValueError("This line should never have been reached")
        def __set__(self, value):
            if self.loc_type == LocType.R2LOC:
                dereference(self._utranspreq._req_r2loc).pickup_timewindow_min = value
            elif self.loc_type == LocType.INT:
                dereference(self._utranspreq._req_int).pickup_timewindow_min = value
            else:
                raise ValueError("This line should never have been reached")



    property pickup_timewindow_max:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self._utranspreq._req_r2loc).pickup_timewindow_max
            elif self.loc_type == LocType.INT:
                return dereference(self._utranspreq._req_int).pickup_timewindow_max
            else:
                raise ValueError("This line should never have been reached")
        def __set__(self, value):
            if self.loc_type == LocType.R2LOC:
                dereference(self._utranspreq._req_r2loc).pickup_timewindow_max = value
            elif self.loc_type == LocType.INT:
                dereference(self._utranspreq._req_int).pickup_timewindow_max = value
            else:
                raise ValueError("This line should never have been reached")


    property delivery_timewindow_min:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self._utranspreq._req_r2loc).delivery_timewindow_min
            elif self.loc_type == LocType.INT:
                return dereference(self._utranspreq._req_int).delivery_timewindow_min
            else:
                raise ValueError("This line should never have been reached")
        def __set__(self, value):
            if self.loc_type == LocType.R2LOC:
                dereference(self._utranspreq._req_r2loc).delivery_timewindow_min = value
            elif self.loc_type == LocType.INT:
                dereference(self._utranspreq._req_int).delivery_timewindow_min = value
            else:
                raise ValueError("This line should never have been reached")



    property delivery_timewindow_max:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self._utranspreq._req_r2loc).delivery_timewindow_max
            elif self.loc_type == LocType.INT:
                return dereference(self._utranspreq._req_int).delivery_timewindow_max
            else:
                raise ValueError("This line should never have been reached")
        def __set__(self, value):
            if self.loc_type == LocType.R2LOC:
                dereference(self._utranspreq._req_r2loc).delivery_timewindow_max = value
            elif self.loc_type == LocType.INT:
                dereference(self._utranspreq._req_int).delivery_timewindow_max = value
            else:
                raise ValueError("This line should never have been reached")

    @staticmethod
    cdef TransportationRequest from_c_r2loc(shared_ptr[CTransportationRequest[R2loc]] creq):
        cdef TransportationRequest req = TransportationRequest.__new__(TransportationRequest)
        req._utranspreq._req_r2loc = creq
        req._ureq._req_r2loc = dynamic_pointer_cast[CRequest[R2loc], CTransportationRequest[R2loc]](creq)
        req.loc_type = LocType.R2LOC
        return req

    @staticmethod
    cdef TransportationRequest from_c_int(shared_ptr[CTransportationRequest[int]] creq):
        cdef TransportationRequest req = TransportationRequest.__new__(TransportationRequest)
        req._utranspreq._req_int = creq
        req._ureq._req_int = dynamic_pointer_cast[CRequest[int], CTransportationRequest[int]](creq)
        req.loc_type = LocType.INT
        return req


    def __dealloc__(self):
        # using unique_ptr's so no deletion
        pass

    def __reduce__(self):
        return self.__class__, \
            (
                self.request_id,
                self.creation_timestamp,
                self.origin,
                self.destination,
                self.pickup_timewindow_min,
                self.pickup_timewindow_max,
                self.delivery_timewindow_min,
                self.delivery_timewindow_max
            )

cdef class InternalRequest(Request):
    """
    Cython equivalent of pure python dataclass :class:`.data_structures.InternalRequest`. See its docstring for details
    of its attributes.

    .. admonition:: Implementation details for developers

       Wraps the C++ templated struct `data_structures_cython.cdata_structures.pxd::InternalRequest[Loc]` into a cython
       extension type. The template parameter `Loc` defines the type of location objects (`Tuple[float, float]` for
       simulations in :math:`\mathbb{R}^2`), `int` for graphs.

       However a cython extension type obviously cannot be templated (at compile time, all possible variations of a template
       must be known). So we will use the Explicit Run-Time Dispatch approach, see this module's docstring or
       :ref:`desc_runtime_dispatch` for details.

       Holds a `shared_ptr` to C++ templated class `InternalRequest<Loc>`. This class can be instantiated from an
       existing `shared_ptr` (only from the cython side and not from the calling python code). See the staticmethods
       `cdef InternalRequest from_c_r2loc` and `cdef InternalRequest from_c_int` for details.
    """
    def __init__(
            self, int request_id, double creation_timestamp,
            location
    ):
        """
        Note
        ----
        We will try to infer `Loc` from what `location` contains.
        """
        if hasattr(location, '__len__') and len(location) == 2:
            # TODO: this inferring of LocType is kludgy. We should have it as an argument of __init__
            # let's assume both origin and destination are Tuple[double, double]
            self.loc_type = LocType.R2LOC
            self._uinternreq._req_r2loc = shared_ptr[CInternalRequest[R2loc]](new CInternalRequest[R2loc](
                request_id, creation_timestamp, location
            ))
            self._ureq._req_r2loc = dynamic_pointer_cast[CRequest[R2loc], CInternalRequest[R2loc]](self._uinternreq._req_r2loc)
        elif isinstance(location, int):
            # let's assume both origin and destination are int
            self.loc_type = LocType.INT
            self._uinternreq._req_int = shared_ptr[CInternalRequest[int]](new CInternalRequest[int](
                request_id, creation_timestamp, location
            ))
            self._ureq._req_int = dynamic_pointer_cast[CRequest[int], CInternalRequest[int]](self._uinternreq._req_int)

        else:
            raise TypeError(f"Cannot handle origin of type {type(location)}")


    def __eq__(self, other: InternalRequest):
        if not isinstance(other, InternalRequest):
            return False
        return self.request_id == other.request_id \
            and self.loc_type == other.loc_type \
            and self.creation_timestamp == other.creation_timestamp \
            and self.location == other.location

    def __repr__(self):
        return f"{self.__class__.__name__}(" + ", ".join(f"{k}={v}" for k, v in self.asdict().items()) + ")"

    def asdict(self):
        return dict(
            request_id=self.request_id,
            creation_timestamp=self.creation_timestamp,
            location=self.location,
        )


    property location:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self._uinternreq._req_r2loc).location
            elif self.loc_type == LocType.INT:
                return dereference(self._uinternreq._req_int).location
            else:
                raise ValueError("This line should never have been reached")
        def __set__(self, value):
            if self.loc_type == LocType.R2LOC:
                dereference(self._uinternreq._req_r2loc).location = value
            elif self.loc_type == LocType.INT:
                dereference(self._uinternreq._req_int).location = value
            else:
                raise ValueError("This line should never have been reached")

    @staticmethod
    cdef InternalRequest from_c_r2loc(shared_ptr[CInternalRequest[R2loc]] creq):
        cdef InternalRequest req = InternalRequest.__new__(InternalRequest)
        req._uinternreq._req_r2loc = creq
        req._ureq._req_r2loc = dynamic_pointer_cast[CRequest[R2loc], CInternalRequest[R2loc]](creq)
        req.loc_type = LocType.R2LOC
        return req

    @staticmethod
    cdef InternalRequest from_c_int(shared_ptr[CInternalRequest[int]] creq):
        cdef InternalRequest req = InternalRequest.__new__(InternalRequest)
        req._uinternreq._req_int = creq
        req._ureq._req_int = dynamic_pointer_cast[CRequest[int], CInternalRequest[int]](creq)
        req.loc_type = LocType.INT
        return req


    def __dealloc__(self):
        # using shared_ptr's so no deletion
        pass


    def __reduce__(self):
        return self.__class__, \
            (
                self.request_id,
                self.creation_timestamp,
                self.location
            )

cdef class Stop:
    """
    Cython equivalent of pure python dataclass :class:`.data_structures.Stop`. See its docstring for details
    of its attributes.

   .. admonition:: Implementation details for developers

    Wraps the C++ templated struct `data_structures_cython.cdata_structures.pxd::Stop[Loc]` into a cython extension type.
    The template parameter `Loc` defines the type of location objects (`Tuple[float, float]` for simulations in
    :math:`\mathbb{R}^2`), `int` for graphs.

    However a cython extension type obviously cannot be templated (at compile time, all possible variations of a template
    must be known). So we will use the Explicit Run-Time Dispatch approach, see this module's docstring or
    :ref:`desc_runtime_dispatch` for details.

    Holds a pointer to C++ templated class `Stop<Loc>`. This class can be instantiated from an existing pointer (only
    from the cython side and not from the calling python code). See the staticmethods `cdef Stop from_c_r2loc` and
    `cdef Stop from_c_int` for details.
    """
    def __cinit__(self):
        pass

    def __init__(
            self,
            location,
            Request request,
            StopAction action,
            double estimated_arrival_time,
            int occupancy_after_servicing,
            double time_window_min=0,
            double time_window_max=inf,
    ):
        """
        Note
        ----
        We will try to infer the `.LocType` from what `location` contains. Do not pass incompatible combinations like
        a `Request` with `LocType=R2LOC` and `location` of type `int`.
        """
        if hasattr(location, '__len__') and len(location) == 2:
            # let's assume both origin and destination are Tuple[double, double]
            self.loc_type = LocType.R2LOC
            self.ustop._stop_r2loc = new CStop[R2loc](
                location, request._ureq._req_r2loc, action, estimated_arrival_time, occupancy_after_servicing,
                time_window_min, time_window_max)
        elif isinstance(location, int):
            # let's assume both origin and destination are int
            self.loc_type = LocType.INT
            self.ustop._stop_int = new CStop[int](
                location, request._ureq._req_int, action, estimated_arrival_time, occupancy_after_servicing,
                time_window_min, time_window_max)
        else:
            raise ValueError("This line should never have been reached")

    def __eq__(self, other: Stop):
        if not isinstance(other, Stop):
            return False
        return self.location == other.location \
            and self.loc_type == other.loc_type \
            and self.request == other.request \
            and self.action == other.action \
            and self.estimated_arrival_time == other.estimated_arrival_time \
            and self.occupancy_after_servicing == other.occupancy_after_servicing \
            and self.time_window_min == other.time_window_min \
            and self.time_window_max == other.time_window_max


    def __deepcopy__(self, *args, **kwargs):
        return Stop(self.location, self.request, self.action, self.estimated_arrival_time,
                    self.occupancy_after_servicing, self.time_window_min, self.time_window_max)


    def __repr__(self):
        return f"{self.__class__.__name__}(" + ", ".join(f"{k}={v}" for k, v in self.asdict().items()) + ")"


    def asdict(self):
        return dict(
            location=self.location,
            request=self.request.asdict(),
            estimated_arrival_time=self.estimated_arrival_time,
            action=self.action,
            time_window_min=self.time_window_min,
            time_window_max=self.time_window_max,
            occupancy_after_servicing=self.occupancy_after_servicing,
        )


    property location:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self.ustop._stop_r2loc).location
            elif self.loc_type == LocType.INT:
                return dereference(self.ustop._stop_int).location
            else:
                raise ValueError("This line should never have been reached")

        def __set__(self, loc):
            if self.loc_type == LocType.R2LOC:
                dereference(self.ustop._stop_r2loc).location = loc
            elif self.loc_type == LocType.INT:
                dereference(self.ustop._stop_int).location = loc
            else:
                raise ValueError("This line should never have been reached")

    property occupancy_after_servicing:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self.ustop._stop_r2loc).occupancy_after_servicing
            elif self.loc_type == LocType.INT:
                return dereference(self.ustop._stop_int).occupancy_after_servicing
            else:
                raise ValueError("This line should never have been reached")

        def __set__(self, occ):
            if self.loc_type == LocType.R2LOC:
                dereference(self.ustop._stop_r2loc).occupancy_after_servicing = occ
            elif self.loc_type == LocType.INT:
                dereference(self.ustop._stop_int).occupancy_after_servicing = occ
            else:
                raise ValueError("This line should never have been reached")

    @property
    def request(self):
        if self.loc_type == LocType.R2LOC:
            if typeid(dereference(dereference(self.ustop._stop_r2loc).request)) == typeid(CTransportationRequest[R2loc]):
                return TransportationRequest.from_c_r2loc(
                    dynamic_pointer_cast[CTransportationRequest[R2loc], CRequest[R2loc]](dereference(self.ustop._stop_r2loc).request))
            elif typeid(dereference(dereference(self.ustop._stop_r2loc).request)) == typeid(CInternalRequest[R2loc]):
                return InternalRequest.from_c_r2loc(
                    dynamic_pointer_cast[CInternalRequest[R2loc], CRequest[R2loc]](dereference(self.ustop._stop_r2loc).request))
            else:
                raise ValueError("This line should never have been reached")
        elif self.loc_type == LocType.INT:
            if typeid(dereference(dereference(self.ustop._stop_int).request)) == typeid(CTransportationRequest[int]):
                return TransportationRequest.from_c_int(
                    dynamic_pointer_cast[CTransportationRequest[int], CRequest[int]](dereference(self.ustop._stop_int).request))
            elif typeid(dereference(dereference(self.ustop._stop_int).request)) == typeid(CInternalRequest[int]):
                return InternalRequest.from_c_int(
                    dynamic_pointer_cast[CInternalRequest[int], CRequest[int]](dereference(self.ustop._stop_int).request))
            else:
                raise ValueError("This line should never have been reached")
        else:
            raise ValueError("This line should never have been reached")

    @property
    def action(self):
        if self.loc_type == LocType.R2LOC:
            return StopAction(dereference(self.ustop._stop_r2loc).action)
        elif self.loc_type == LocType.INT:
            return StopAction(dereference(self.ustop._stop_int).action)
        else:
            raise ValueError("This line should never have been reached")

    property estimated_arrival_time:
        def __get__(self):
            if self.loc_type == LocType.R2LOC:
                return dereference(self.ustop._stop_r2loc).estimated_arrival_time
            elif self.loc_type == LocType.INT:
                return dereference(self.ustop._stop_int).estimated_arrival_time
            else:
                raise ValueError("This line should never have been reached")

        def __set__(self, estimated_arrival_time):
            if self.loc_type == LocType.R2LOC:
                dereference(self.ustop._stop_r2loc).estimated_arrival_time = estimated_arrival_time
            elif self.loc_type == LocType.INT:
                dereference(self.ustop._stop_int).estimated_arrival_time = estimated_arrival_time
            else:
                raise ValueError("This line should never have been reached")

    @property
    def time_window_min(self):
        if self.loc_type == LocType.R2LOC:
            return dereference(self.ustop._stop_r2loc).time_window_min
        elif self.loc_type == LocType.INT:
            return dereference(self.ustop._stop_int).time_window_min
        else:
            raise ValueError("This line should never have been reached")

    @property
    def time_window_max(self):
        if self.loc_type == LocType.R2LOC:
            return dereference(self.ustop._stop_r2loc).time_window_max
        elif self.loc_type == LocType.INT:
            return dereference(self.ustop._stop_int).time_window_max
        else:
            raise ValueError("This line should never have been reached")



    @staticmethod
    cdef Stop from_c_r2loc(CStop[R2loc] *cstop):
        cdef Stop stop = Stop.__new__(Stop)
        stop.ptr_owner = False
        stop.ustop._stop_r2loc = cstop
        stop.loc_type = LocType.R2LOC
        return stop

    @staticmethod
    cdef Stop from_c_int(CStop[int] *cstop):
        cdef Stop stop = Stop.__new__(Stop)
        stop.ptr_owner = False
        stop.ustop._stop_int = cstop
        stop.loc_type = LocType.INT
        return stop

    def __reduce__(self):
        return self.__class__, \
            (
                self.location,
                self.request,
                self.action.value,
                self.estimated_arrival_time,
                self.occupancy_after_servicing,
                self.time_window_min,
                self.time_window_max,
            )

cdef class Stoplist:
    """
    The cythonic equivalent of `.data_structures.Stoplist`, which is just a python list of `.data_structures.Stop`.

    Note
    ----
    An instance of this class doies not support most of the functionalities of python lists, e.g. slicing. Please take a
    look at the available methods. In particular, use `stoplist.remove_nth_elem(int)` instead of `stoplist.pop(int)` or
    `del stoplist[i]`.


    .. admonition:: Implementation details for developers

       Wraps a C++ vector of `data_structures_cython.cdata_structures.pxd::Stop[Loc]` into a cython extension type.
       The template parameter `Loc` defines the type of location objects (`Tuple[float, float]` for simulations in
       :math:`\mathbb{R}^2`), `int` for graphs.

       However a cython extension type obviously cannot be templated (at compile time, all possible variations of a template
       must be known). So we will use the Explicit Run-Time Dispatch approach, see this module's docstring or
       :ref:`desc_runtime_dispatch` for details.

       This class can be instantiated from an existing C++ `vector<Stop>`, only from the cython side and not from the calling
       python code. See the staticmethods `Stoplist from_c_r2loc` and `Stoplist from_c_int` for details.
    """
    # TODO: May need to allow nice ways of converting a Stoplist to python lists or similar. Use case: calling code
    # doing post optimization
    def __init__(self, python_stoplist, loc_type):
        """
        Parameters
        ----------
        python_stoplist
            A python list of :class:`Stop` objects
        loc_type
            The type of location objects, as defined in the enum `.LocType`

        Note
        ----
        Do **not** pass a list of pure pythonic :class:`Stop<.data_structures.Stop>` objects to `__init__`, this will
        lead to a crash. Only pass a list of cython extension :class:`.Stop` objects.

        The `Stop` objects are **copied** here, therefore the original `Stop` objects can safely be garbage collected/
        otherwise deleted.
        """
        self.loc_type = loc_type

        if self.loc_type == LocType.R2LOC:
            self.ustoplist._stoplist_r2loc = vector[CStop[R2loc]](0)
        elif self.loc_type == LocType.INT:
            self.ustoplist._stoplist_int = vector[CStop[int]](0)
        else:
            raise ValueError(f"This line should never have been reached: {type(loc_type)}")

        if loc_type == LocType.R2LOC:
            for py_s in python_stoplist:
                self.ustoplist._stoplist_r2loc.push_back(dereference((<Stop> py_s).ustop._stop_r2loc))
        elif loc_type == LocType.INT:
            for py_s in python_stoplist:
                self.ustoplist._stoplist_int.push_back(dereference((<Stop> py_s).ustop._stop_int))
        else:
            raise ValueError("This line should never have been reached")
        logger.info("Created Stoplist")

    def __getitem__(self, i):
        """
        Works like `list.__getitem__`. Allows indexing by negative integers.
        """
        len_ = self.__len__()
        if 0 <= i < len_:
            i = i
        elif -len_ <= i < 0:
            i = len_+i
        else:
            raise IndexError(f"list index {i} out of range")
        if self.loc_type == LocType.R2LOC:
            return Stop.from_c_r2loc(&self.ustoplist._stoplist_r2loc[i])
        elif self.loc_type == LocType.INT:
            return Stop.from_c_int(&self.ustoplist._stoplist_int[i])
        else:
            raise ValueError("This line should never have been reached")

    def __len__(self):
        """
        Works like `list.__len__`.
        """
        if self.loc_type == LocType.R2LOC:
            return <int> self.ustoplist._stoplist_r2loc.size()
        elif self.loc_type == LocType.INT:
            return <int> self.ustoplist._stoplist_int.size()
        else:
            raise ValueError("This line should never have been reached")

    def remove_nth_elem(self, int n):
        """
        Removes the nth element. Does not return anything.
        """
        if self.loc_type == LocType.R2LOC:
            self.ustoplist._stoplist_r2loc.erase(
                self.ustoplist._stoplist_r2loc.begin()+n
            )
        elif self.loc_type == LocType.INT:
            self.ustoplist._stoplist_int.erase(
                self.ustoplist._stoplist_int.begin()+n
            )
        else:
            raise ValueError("This line should never have been reached")

    @staticmethod
    cdef Stoplist from_c_r2loc(vector[CStop[R2loc]] cstoplist):
        cdef Stoplist stoplist = Stoplist.__new__(Stoplist) # Calling __new__ bypasses __init__, see
        # https://cython.readthedocs.io/en/latest/src/userguide/extension_types.html#instantiation-from-existing-c-c-pointers
        stoplist.loc_type = LocType.R2LOC
        stoplist.ustoplist._stoplist_r2loc = cstoplist
        return stoplist

    @staticmethod
    cdef Stoplist from_c_int(vector[CStop[int]] cstoplist):
        cdef Stoplist stoplist = Stoplist.__new__(Stoplist) # Calling __new__ bypasses __init__, see
        # https://cython.readthedocs.io/en/latest/src/userguide/extension_types.html#instantiation-from-existing-c-c-pointers
        stoplist.loc_type = LocType.INT
        stoplist.ustoplist._stoplist_int = cstoplist
        return stoplist

    def __repr__(self):
        if self.loc_type == LocType.R2LOC:
            return f"[{', '.join(repr(Stop.from_c_r2loc(&s)) for s in self.ustoplist._stoplist_r2loc)}]"
        elif self.loc_type == LocType.INT:
            return f"[{', '.join(repr(Stop.from_c_int(&s)) for s in self.ustoplist._stoplist_int)}]"
        else:
            raise ValueError("This line should never have been reached")


    def to_pys(self):
        logger.debug("Converting cython Stoplist to python list")
        if self.loc_type == LocType.R2LOC:
            # According to cython docs, the following should have worked:
            # return list([Stop.from_c_r2loc(&s) for s in self.ustoplist._stoplist_r2loc])
            # However, it doesn't. So we need the more verbose loop:
            res = []
            for idx in range(self.ustoplist._stoplist_r2loc.size()):
                res.append(Stop.from_c_r2loc(&self.ustoplist._stoplist_r2loc[idx]))
            return res
        elif self.loc_type == LocType.INT:
            res = []
            for idx in range(self.ustoplist._stoplist_int.size()):
                res.append(Stop.from_c_int(&self.ustoplist._stoplist_int[idx]))
            return res
        else:
            raise ValueError("This line should never have been reached")

    def __reduce__(self):
        return self.__class__, (self.to_pys(), self.loc_type)

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        return all(self[i]==other[i] for i in range(len(self)))



from __future__ import annotations

from collections import defaultdict

import functools as ft
import itertools as it
import operator as op

import numpy as np


from abc import ABC, abstractmethod
from typing import (
    Dict,
    SupportsFloat,
    Iterator,
    Union,
    Tuple,
    Iterable,
    Type,
)

import logging

from .util import supress_stoplist_extraction_warning

logger = logging.getLogger(__name__)

from ridepy.data_structures import (
    Stop,
    Request,
    TransportationRequest as pyTransportationRequest,
    InternalRequest as pyInternalRequest,
    SingleVehicleSolution,
    Dispatcher,
    StopAction,
    InternalRequest,
    TransportationRequest,
)
from .events import (
    Event,
    RequestEvent,
    StopEvent,
)
from ridepy.data_structures_cython import (
    TransportationRequest as CyTransportationRequest,
    InternalRequest as CyInternalRequest,
    StopAction as CyStopAction,
    Stop as CyStop,
)

from ridepy.vehicle_state import VehicleState
from ridepy.vehicle_state_cython import VehicleState as CyVehicleState

from ridepy.data_structures import TransportSpace
from ridepy.util.spaces_cython import TransportSpace as CyTransportSpace


class FleetState(ABC):
    """
    Provides the interface for running the whole simulation.
    The fleet state is stored as a dictionary of (vehicle_id, VehicleState) pairs.
    Exposes the methods fast_forward, handle_request and simulate.

    To gain maximum computational efficiency, one can:

    * Subclass VehicleState and implement the main methods in Cython.
    * Implement fast_forward and handle_request using distributed computing e.g. MPI
    """

    def __init__(
        self,
        *,
        initial_locations: Union[Dict[int, int], Dict[int, Tuple[float, ...]]],
        vehicle_state_class: Union[Type[VehicleState], Type[CyVehicleState]],
        space: Union[TransportSpace, CyTransportSpace],
        dispatcher: Dispatcher,
        seat_capacities: Union[int, Dict[int, int]],
    ):
        """
        Create a `FleetState`, holding a number of `VehicleState` objects.

        Parameters
        ----------
        initial_locations
            Dictionary with vehicle ids as keys and initial locations as values.
        vehicle_state_class
            The vehicle state class to be used. Can be either `.vehicle_state.VehicleState` (pure pythonic) or
            `.vehicle_state_cython.VehicleState` (implemented in cython).
        space
            The `.data_structures.TransportSpace` (e.g. Euclidean2D, Graph) in which the simulation will be run.
        dispatcher
            The dispatching algorithm that maps a (stoplist, TransportationRequest) pair into a cost and new stoplist.
            See :doc:`introduction` for more details.
        seat_capacities
            Integers denoting the maximum number of persons a vehicle can hold. Either a dictionary with vehicle ids as
            keys or a positive integer (Causes each vehicle to have the same capacity).

        Note
        ----
        * Explain the graph nodes have in notes and why.
        * define LocType in a central place and motivate why we have it.
        """
        if issubclass(vehicle_state_class, VehicleState):
            StopCls = Stop
            StopActionCls = StopAction
            InternalRequestCls = InternalRequest
            TransportationRequestCls = TransportationRequest
            assert isinstance(space, TransportSpace), "unsuitable transport space"
            logger.debug(
                f"Creating FleetState with vehicle state class {vehicle_state_class}"
            )
        elif issubclass(vehicle_state_class, CyVehicleState):
            StopCls = CyStop
            StopActionCls = CyStopAction
            InternalRequestCls = CyInternalRequest
            TransportationRequestCls = CyTransportationRequest
            assert isinstance(space, CyTransportSpace), "unsuitable transport space"
            logger.debug(
                f"Creating FleetState with vehicle state class {vehicle_state_class}"
            )
        else:
            raise TypeError(f"Unknown VehicleStateCls {type(vehicle_state_class)}")

        if isinstance(seat_capacities, Dict):
            assert set(initial_locations) == set(
                seat_capacities
            ), "vehicle_ids in seat_capacities and initial_stoplists must match"
        elif isinstance(seat_capacities, int):
            seat_capacities = defaultdict(lambda x=seat_capacities: x)
        else:
            raise TypeError(f"seat_capacities must be either dict or int")

        self.vehicle_state_class = vehicle_state_class
        self.dispatcher = dispatcher
        self.space = space

        assert initial_locations, "No initial locations supplied."
        for initial_location in initial_locations.values():
            # note that NumPy's dimensions start from 0
            assert space.n_dim == np.ndim(initial_location) + 1, (
                f"Dimension mismatch: Initial location {initial_location} of "
                f"dimensionality {np.ndim(initial_location) + 1} supplied, "
                f"but space has {space.n_dim} dimensions."
            )

        self.fleet: Dict[int, VehicleState] = {
            vehicle_id: vehicle_state_class(
                vehicle_id=vehicle_id,
                initial_stoplist=[
                    StopCls(
                        location=initial_locations[vehicle_id],
                        request=InternalRequestCls(
                            request_id=-1,
                            creation_timestamp=0,
                            location=initial_locations[vehicle_id],
                        ),
                        action=StopActionCls.internal,
                        estimated_arrival_time=0,
                        occupancy_after_servicing=0,
                        time_window_min=0,
                        time_window_max=np.inf,
                    )
                ],
                space=space,
                dispatcher=self.dispatcher,
                seat_capacity=seat_capacities[vehicle_id],
            )
            for vehicle_id in initial_locations.keys()
        }

    @classmethod
    def from_fleet(
        cls,
        *,
        fleet: Dict[int, VehicleState],
        space: Union[TransportSpace, CyTransportSpace],
        dispatcher: Dispatcher,
        validate: bool = True,
    ):
        """
        Create a `FleetState` from a dictionary of `VehicleState` objects, keyed by an integer `vehicle_id`.

        Parameters
        ----------
        fleet
            Initial stoplists *must* contain current position element (CPE) stops with
            `StopAction.internal` and `InternalRequest` with `request_id==-1` as their first entries.
        space
            TransportSpace to operate on
        dispatcher
            dispatcher to use to assign requests to vehicles or reject them.
        validate
            If true, try to figure out whether something is off. If not, raise.

        Returns
        -------

        """
        self = super().__new__(cls)
        self.dispatcher = dispatcher
        self.fleet = fleet
        self.space = space

        if validate:
            VehicleStateCls = type(next(iter(self.fleet.values())))
            assert all(
                isinstance(vehicle_state, VehicleStateCls)
                for vehicle_state in self.fleet.values()
            ), "fleet inhomogeneous"

            if issubclass(VehicleStateCls, VehicleState):
                StopCls = Stop
                StopActionCls = StopAction
                InternalRequestCls = InternalRequest
                TransportationRequestCls = TransportationRequest
                assert isinstance(space, TransportSpace), "unsuitable transport space"
            elif issubclass(VehicleStateCls, CyVehicleState):
                StopCls = CyStop
                StopActionCls = CyStopAction
                InternalRequestCls = CyInternalRequest
                TransportationRequestCls = CyTransportationRequest
                assert isinstance(space, CyTransportSpace), "unsuitable transport space"
            else:
                raise TypeError(f"Unknown VehicleStateCls {type(VehicleStateCls)}")

            for vehicle_state in self.fleet.values():
                assert (
                    vehicle_state.stoplist[0].request.request_id == -1
                ), "malformed CPE: request_id must be -1"
                assert (
                    vehicle_state.stoplist[0].action == StopActionCls.internal
                ), "malformed CPE: action must be 'internal'"
                assert isinstance(
                    vehicle_state.stoplist[0].request, InternalRequestCls
                ), "malformed CPE: request type must be 'internal'"
                assert all(isinstance(stop, StopCls) for stop in vehicle_state.stoplist)
                assert all(
                    isinstance(
                        stop.request, (InternalRequestCls, TransportationRequestCls)
                    )
                    for stop in vehicle_state.stoplist
                )

        return self

    def simulate(
        self, requests: Iterator[Request], t_cutoff: float = np.inf
    ) -> Iterator[Event]:
        """
        Run a simulation.

        Parameters
        ----------
        requests
            `Iterator` that supplies incoming requests.
        t_cutoff
            optional cutoff time after which the simulation is forcefully ended,
            disregarding any remaining stops or requests.

        Returns
        -------
        Iterator of events that have been emitted during the simulation.

        Note
        ----
        Because of lazy evaluation the returned iterator must be exhausted
        for the simulation to be actually be performed.
        """

        self.t = 0

        with supress_stoplist_extraction_warning():
            for vehicle_id, fleet_state in self.fleet.items():
                yield {
                    "event_type": "VehicleStateBeginEvent",
                    "vehicle_id": vehicle_id,
                    "timestamp": self.t,
                    "location": fleet_state.stoplist[0].location,
                    "request_id": -100,
                }

        for n_req, request in enumerate(requests):
            # advance clock to req_epoch
            self.t = request.creation_timestamp

            if self.t > t_cutoff:
                break

            # Visit all the stops upto req_epoch
            yield from self.fast_forward(self.t)

            if isinstance(request, (TransportationRequest, CyTransportationRequest)):
                yield {
                    "event_type": "RequestSubmissionEvent",
                    "request_id": request.request_id,
                    "timestamp": self.t,
                    "origin": request.origin,
                    "destination": request.destination,
                    "pickup_timewindow_min": request.pickup_timewindow_min,
                    "pickup_timewindow_max": request.pickup_timewindow_max,
                    "delivery_timewindow_min": request.delivery_timewindow_min,
                    "delivery_timewindow_max": request.delivery_timewindow_max,
                }

            # handle the current request
            if isinstance(request, (pyTransportationRequest, CyTransportationRequest)):
                yield self.handle_transportation_request(request)
            elif isinstance(request, (pyInternalRequest, CyInternalRequest)):
                yield self.handle_internal_request(request)
            else:
                raise NotImplementedError(f"Unknown request type: {type(request)}")
            logger.info(f"Handled request # {n_req}")

        with supress_stoplist_extraction_warning():
            self.t = min(
                t_cutoff,
                max(
                    vehicle.stoplist[-1].estimated_arrival_time
                    for vehicle in self.fleet.values()
                ),
            )

        # service all remaining stops
        yield from self.fast_forward(self.t)

        with supress_stoplist_extraction_warning():
            for vehicle_id, fleet_state in self.fleet.items():
                yield {
                    "event_type": "VehicleStateEndEvent",
                    "vehicle_id": vehicle_id,
                    "timestamp": self.t,
                    "location": fleet_state.stoplist[0].location,
                    "request_id": -200,
                }

    @abstractmethod
    def fast_forward(self, t: SupportsFloat) -> Iterator[StopEvent]:
        """
        Advance the simulator's state in time from the previous time :math:`t'`
        to the new time :math:`t`, with :math:`t >= t'`.
        E.g. vehicle locations may change and vehicle stops may be serviced.
        The latter will emit `.StopEvent` s which are returned.

        Parameters
        ----------
        t
            Time to advance to.

        Returns
        -------
            Iterator of stop events. May be empty if no stop is serviced.
        """
        ...

    def handle_transportation_request(
        self, req: pyTransportationRequest
    ) -> RequestEvent:
        """
        Handle a request by mapping the request and the fleet state onto a request response,
        modifying the fleet state in-place.

        This method can be implemented in child classes using various parallelization techniques like
        `multiprocessing`, `OpenMPI` or `dask`.

        Parameters
        ----------
        req
            Request to handle.

        Returns
        -------
            An `.Event`.
        """
        ...

    @abstractmethod
    def handle_internal_request(self, req: pyInternalRequest) -> RequestEvent:
        """

        Parameters
        ----------
        req
            request to handle.

        Returns
        -------
            An `.Event`.
        """
        ...

    def _apply_request_solution(
        self, req, all_solutions: Iterable[SingleVehicleSolution]
    ) -> RequestEvent:
        """
        Given a request and a bunch of solutions, pick the one with the minimum cost and apply it,
        thereby changing the stoplist of the chosen vehicle and emitting an RequestAcceptanceEvent.
        If the minimum cost is infinite, RequestRejectionEvent is returned.

        Parameters
        ----------
        req
            request to handle.
        all_solutions
            a dictionary mapping a vehicle_id to a `.SingleVehicleSolution`.

        Returns
        -------
            Either a `.RequestAcceptanceEvent`, or a `.RequestRejectionEvent` if no suitable vehicle could be found.

        Note
        ----
            Modifies the `.VehicleState` of the vehicle with the least cost inplace.

        """
        (
            best_vehicle,
            min_cost,
            (
                pickup_timewindow_min,
                pickup_timewindow_max,
                delivery_timewindow_min,
                delivery_timewindow_max,
            ),
        ) = min(all_solutions, key=op.itemgetter(1))
        logger.debug(f"best vehicle: {best_vehicle}, at min_cost={min_cost}")
        if min_cost == np.inf:  # no solution was found
            return {
                "event_type": "RequestRejectionEvent",
                "timestamp": self.t,
                "request_id": req.request_id,
            }
        else:
            # modify the best vehicle's stoplist
            self.fleet[best_vehicle].select_new_stoplist()

            return {
                "event_type": "RequestAcceptanceEvent",
                "timestamp": self.t,
                "request_id": req.request_id,
                "origin": req.origin,
                "destination": req.destination,
                "pickup_timewindow_min": pickup_timewindow_min,
                "pickup_timewindow_max": pickup_timewindow_max,
                "delivery_timewindow_min": delivery_timewindow_min,
                "delivery_timewindow_max": delivery_timewindow_max,
            }


class SlowSimpleFleetState(FleetState):
    def fast_forward(self, t: float):
        events = (
            vehicle_state.fast_forward_time(t) for vehicle_state in self.fleet.values()
        )

        # no need to swap the old stoplists of each vehicle with the new (fast-forwarded)
        # stoplists, because vehicle_state_class.fast_forward did that already.

        return sorted(it.chain.from_iterable(events), key=op.itemgetter("timestamp"))

    def handle_transportation_request(
        self, req: pyTransportationRequest
    ) -> RequestEvent:
        logger.debug(f"Handling Request: {req}")

        if req.origin == req.destination:
            return {
                "event_type": "RequestRejectionEvent",
                "timestamp": self.t,
                "request_id": req.request_id,
            }

        return self._apply_request_solution(
            req,
            map(
                ft.partial(
                    self.vehicle_state_class.handle_transportation_request_single_vehicle,
                    request=req,
                ),
                self.fleet.values(),
            ),
        )

    def handle_internal_request(self, req: pyInternalRequest) -> RequestEvent: ...



import random
from abc import (
    ABC,
    abstractmethod,
    abstractproperty,
    abstractclassmethod,
    abstractstaticmethod,
)
from typing import Optional, Union

import numpy as np

from ridepy.data_structures import (
    TransportSpace,
    TransportationRequest,
    RequestGenerator,
)

from ridepy.util.spaces import Euclidean2D
from ridepy.util.spaces_cython import Euclidean2D as CyEuclidean2D


class RandomRequestGenerator(RequestGenerator):
    """
    Generates random requests in the chosen transport space with a certain rate.
    The timewindows for the request are configurable.

    Note:
        The pickup and dropoff timewindows are calculated as follows:
            1. request_pickup_time < request.creation_timestamp + max_pickup_delay
            2. request_delivery_time < request.creation_timestamp + direct_travel_time \
                + max_delivery_delay_abs
            3. request_delivery_time < request.creation_timestamp + direct_travel_time*(1+max_delivery_delay_rel)
    """

    def __init__(
        self,
        *,
        space: TransportSpace,
        rate=1,
        seed=42,
        pickup_timewindow_offset=0,
        request_class=TransportationRequest,
        max_pickup_delay: float = np.inf,
        max_delivery_delay_abs: float = np.inf,
        max_delivery_delay_rel: float = np.inf,
    ):
        """

        Parameters
        ----------
        space
            the TransportSpace in which the requests will be generated.
        rate
            the rate of requests per time unit
        seed
            the random seed
        request_class
            the generated requests will be instances of this class. Needed to generate pythonic or cythonic requests at will.
        pickup_timewindow_offset
            Each request's pickup_timewindow_min will be this much from the creation
            timestamp.
        max_pickup_delay
            The maximum delay between the request creation timestamp and the pickup
            time.
        max_delivery_delay_abs
            The maximum absolute delay between the request creation timestamp and the
            delivery time.
        max_delivery_delay_rel
            The maximum **delay** between the request creation timestamp and the delivery
            time relative to the direct actual direct travel time. Example: if the
            direct travel time is 10 minutes and ``max_delivery_delay_rel`` is 0.5,
            the delivery time will be at most 15 minutes after request creation.
            I.e., delta_max = 1 + max_delivery_delay_rel.
        """
        if seed is not None:
            np.random.seed(seed)
            random.seed(seed)

        self.transport_space = space
        self.rate = rate
        self.request_class = request_class
        self.pickup_timewindow_offset = pickup_timewindow_offset
        self.max_pickup_delay = max_pickup_delay
        self.max_delivery_delay_abs = max_delivery_delay_abs
        self.max_delivery_delay_rel = max_delivery_delay_rel

    def __iter__(self):
        self.now = 0
        self.request_index = -1
        return self

    def __next__(self):
        self.now += np.random.exponential(1 / self.rate)
        self.request_index += 1

        while True:
            origin = self.transport_space.random_point()
            destination = self.transport_space.random_point()
            if origin != destination:
                break

        direct_travel_time = self.transport_space.t(origin, destination)
        pickup_lbound = self.now + self.pickup_timewindow_offset
        pickup_ubound = pickup_lbound + self.max_pickup_delay
        delivery_ubound = (
            pickup_lbound
            + direct_travel_time
            + min(
                self.max_delivery_delay_abs,
                self.max_delivery_delay_rel * direct_travel_time,
            )
        )

        return self.request_class(
            request_id=self.request_index,
            creation_timestamp=self.now,
            origin=origin,
            destination=destination,
            pickup_timewindow_min=pickup_lbound,
            pickup_timewindow_max=pickup_ubound,
            delivery_timewindow_min=pickup_lbound,
            delivery_timewindow_max=delivery_ubound,
        )



import random
import copy
import warnings

from typing import List, Tuple, Union, Any, Iterator, Sequence

import numpy as np
import operator as op
import math as m
import networkx as nx
import itertools as it
from scipy.spatial import distance as spd

from ridepy.data_structures import TransportSpace, ID, LocType
from ridepy.util import smartVectorize, make_repr


class Euclidean(TransportSpace):
    """
    n-dimensional Euclidean space with constant velocity.
    """

    def __init__(
        self,
        n_dim: int = 1,
        coord_range: List[Tuple[Union[int, float], Union[int, float]]] = None,
        velocity: float = 1,
    ):
        """
        Initialize n-dimensional Euclidean space with constant velocity.

        Parameters
        ----------
        n_dim
            number of dimensions
        coord_range
            coordinate range of the space as a list of 2-tuples (x_i,min, x_i,max)
            where x_i represents the ith dimension.
        velocity
            constant scaling factor as discriminator between distance and travel time
        """
        self.n_dim = n_dim
        self.velocity = velocity

        if coord_range is not None:
            assert len(coord_range) == n_dim, (
                "Number of desired dimensions must "
                "match the number of coord range pairs given"
            )
            self.coord_range = coord_range
        else:
            self.coord_range = [(0, 1)] * n_dim

    @smartVectorize
    def d(self, u, v):
        if self.n_dim == 1:
            return abs(v - u)
        else:
            return spd.euclidean(u, v)

    def t(self, u, v):
        return self.d(u, v) / self.velocity

    def _coord_sub(self, u, v):
        if self.n_dim == 1:
            return u - v
        else:
            return map(op.sub, u, v)

    def _coord_mul(self, u, k):
        if self.n_dim == 1:
            return u * k
        else:
            return map(op.mul, u, it.repeat(k))

    def interp_dist(self, u, v, dist_to_dest):
        return (
            self._coord_sub(
                v, self._coord_mul(self._coord_sub(v, u), dist_to_dest / self.d(u, v))
            ),
            0,
        )

    def interp_time(self, u, v, time_to_dest):
        return (
            self._coord_sub(
                v, self._coord_mul(self._coord_sub(v, u), time_to_dest / self.t(u, v))
            ),
            0,
        )

    def random_point(self):
        return tuple(random.uniform(a, b) for a, b in self.coord_range)

    def asdict(self):
        return dict(
            n_dim=self.n_dim, coord_range=self.coord_range, velocity=self.velocity
        )

    def __repr__(self):
        return make_repr("Euclidean", self.asdict())


class Euclidean1D(Euclidean):
    def __init__(
        self,
        coord_range: List[Tuple[Union[int, float], Union[int, float]]] = None,
        velocity: float = 1,
    ):
        super().__init__(n_dim=1, coord_range=coord_range, velocity=velocity)

    def interp_dist(self, u, v, dist_to_dest):
        return v - (v - u) * dist_to_dest / self.d(u, v), 0

    def interp_time(self, u, v, time_to_dest):
        return v - (v - u) * time_to_dest / self.t(u, v), 0

    @smartVectorize
    def d(self, u, v):
        return abs(v - u)

    def random_point(self):
        return random.uniform(self.coord_range[0][0], self.coord_range[0][1])

    def asdict(self):
        return dict(coord_range=self.coord_range, velocity=self.velocity)

    def __repr__(self):
        return make_repr("Euclidean1D", self.asdict())


class Euclidean2D(Euclidean):
    def __init__(
        self,
        coord_range: List[Tuple[Union[int, float], Union[int, float]]] = None,
        velocity: float = 1,
    ):
        super().__init__(n_dim=2, coord_range=coord_range, velocity=velocity)
        self.loc_type = LocType.R2LOC

    def _coord_sub(self, u, v):
        return u[0] - v[0], u[1] - v[1]

    def _coord_mul(self, u, k):
        return u[0] * k, u[1] * k

    @smartVectorize
    def d(self, u, v):
        return m.sqrt(m.pow(v[0] - u[0], 2) + m.pow(v[1] - u[1], 2))

    def asdict(self):
        return dict(coord_range=self.coord_range, velocity=self.velocity)

    def __repr__(self):
        return make_repr("Euclidean2D", self.asdict())


class Manhattan2D(TransportSpace):
    """
    :math:`\mathbb{R}^2` with :math:`L_1`-induced metric (Manhattan).
    """

    def __init__(
        self,
        coord_range: List[Tuple[Union[int, float], Union[int, float]]] = None,
        velocity: float = 1,
    ):
        """
        Parameters
        ----------
        coord_range
            coordinate range of the space as a list of 2-tuples (x_i,min, x_i,max)
            where x_i represents the ith dimension.
        velocity
            constant scaling factor as discriminator between distance and travel time
        """
        self.n_dim = 2
        self.velocity = velocity
        self.loc_type = LocType.R2LOC

        if coord_range is not None:
            assert len(coord_range) == 2, (
                "Number of desired dimensions must "
                "match the number of coord range pairs given"
            )
            self.coord_range = coord_range
        else:
            self.coord_range = [(0, 1)] * 2

    @smartVectorize
    def d(self, u, v):
        return abs(u[0] - v[0]) + abs(u[1] - v[1])

    def t(self, u, v):
        return self.d(u, v) / self.velocity

    def _coord_sub(self, u, v):
        return map(op.sub, u, v)

    def _coord_mul(self, u, k):
        return map(op.mul, u, it.repeat(k))

    def interp_dist(self, u, v, dist_to_dest):
        return (
            self._coord_sub(
                v, self._coord_mul(self._coord_sub(v, u), dist_to_dest / self.d(u, v))
            ),
            0,
        )

    def interp_time(self, u, v, time_to_dest):
        return (
            self._coord_sub(
                v, self._coord_mul(self._coord_sub(v, u), time_to_dest / self.t(u, v))
            ),
            0,
        )

    def random_point(self):
        return tuple(random.uniform(a, b) for a, b in self.coord_range)

    def asdict(self):
        return dict(coord_range=self.coord_range, velocity=self.velocity)

    def __repr__(self):
        return make_repr("Manhattan", self.asdict())


class Graph(TransportSpace):
    """
    A location is identified with a one-dimensional coordinate, namely the node index.
    """

    n_dim = 1
    loc_type = LocType.INT

    def _update_distance_cache(self):
        (
            self._predecessors,
            self._distances,
        ) = nx.floyd_warshall_predecessor_and_distance(self.G, "distance")

    def __init__(
        self,
        vertices: Sequence[int],
        edges: Sequence[Tuple[int, int]],
        weights: Union[None, float, Sequence[float]] = None,
        velocity: float = 1,
    ):
        """
        Weighted undirected graph with integer vertex labels

        Parameters
        ----------
        vertices
            sequence of vertices
        edges
            sequence of edge tuples
        weights
            Edge weights.
            - if None is supplied, the resulting graph is unweighted (unit edge length)
            - if a single float is supplied, every edge length will be equal to this number
            - if a sequence is supplied, this will be mapped onto the edge sequence
        velocity
            constant velocity to compute travel time, optional.
        """
        self.G = nx.Graph()
        self.G.add_nodes_from(vertices)

        if weights is None:
            weights = 1

        if isinstance(weights, (int, float)):
            weights = it.repeat(float(weights))

        self.G.add_edges_from(
            (u, v, {"distance": w}) for (u, v), w in zip(edges, weights)
        )

        self.loc_type = LocType.INT
        self.velocity = velocity
        self._update_distance_cache()

    @staticmethod
    def _prepare_copy_of_nx_graph(*, G, make_attribute_distance, directed=False):
        if not directed:
            graph_class = nx.Graph
        else:
            graph_class = nx.DiGraph

        if not all(
            isinstance(u, int) for u in random.sample(list(G.nodes()), k=min(5, len(G)))
        ):
            warnings.warn(
                "Heuristic determined non-int node labels. Converting to int, "
                "keeping original labels as attribute.",
                UserWarning,
            )

            G = nx.relabel.convert_node_labels_to_integers(
                G, label_attribute="original_label"
            )
        else:
            # create a copy as we don't want to modify the original graph
            G = copy.deepcopy(G)

        # as we are keeping the graph instance, make sure it's right
        if not isinstance(G, graph_class):
            raise TypeError(f"Must supply {graph_class.__name__}, not {type(G)}")
        elif (directed and not G.is_directed()) or (not directed and G.is_directed()):
            raise TypeError(f"Must supply {'' if directed else 'un'}directed graph")

        # making another attribute the distance
        if make_attribute_distance is None:
            nx.set_edge_attributes(G, 1, name="distance")
        elif make_attribute_distance != "distance":
            # if 'distance' already exists, we raise
            if nx.get_edge_attributes(G, "distance"):
                raise ValueError(
                    "'distance' already exists as edge attribute, won't overwrite"
                )
            # otherwise rename
            else:
                nx.set_edge_attributes(
                    G,
                    nx.get_edge_attributes(G, make_attribute_distance),
                    name="distance",
                )
        else:
            if len(nx.get_edge_attributes(G, "distance")) != G.number_of_edges():
                raise ValueError(
                    "Was told to use 'distance' as distance edge attribute. "
                    "All edges must have the distance weight specified."
                )

        return G

    @classmethod
    def from_nx(
        cls,
        G: nx.Graph,
        velocity: float = 1,
        make_attribute_distance: str = "distance",
    ):
        """
        Create a graph space from NetworkX graph.

        Parameters
        ----------
        G
            the networkx.Graph instance, will be deepcopied
        velocity
            velocity to use for travel time computation
        make_attribute_distance
            attribute to rename to "distance" and use as such
            If None is supplied, the resulting graph is unweighted (unit edge length).

        Returns
        -------
        graph space instance
        """
        self = cls.__new__(cls)
        self.G = cls._prepare_copy_of_nx_graph(
            G=G, make_attribute_distance=make_attribute_distance, directed=False
        )
        self.velocity = velocity
        self._update_distance_cache()
        return self

    @smartVectorize
    def d(self, u, v):
        return self._distances[u][v]

    def t(self, u, v) -> Union[int, float]:
        return self.d(u, v) / self.velocity

    def interp_dist(self, u, v, dist_to_dest):
        """

        Parameters
        ----------
        u
        v
        dist_to_dest

        Returns
        -------
        next_node
        jump_distance

        """

        if u == v:
            return v, 0

        next_node = v
        while next_node is not u:
            predecessor = self._predecessors[u][next_node]
            predecessor_dist = self.d(predecessor, v)
            if predecessor_dist >= dist_to_dest:
                break
            next_node = predecessor

        if predecessor_dist > dist_to_dest:
            return next_node, dist_to_dest - self.d(next_node, v)
        else:
            return predecessor, 0

    def interp_time(self, u, v, time_to_dest):
        """

        Parameters
        ----------
        u
        v
        time_to_dest

        Returns
        -------
        next_node
        jump_time

        """
        next_node, jump_dist = self.interp_dist(
            u, v, dist_to_dest=time_to_dest * self.velocity
        )
        return next_node, jump_dist / self.velocity

    def shortest_path_vertex_sequence(self, u, v) -> List[ID]:
        seq = [v]
        if u != v:
            next_node = v
            while next_node is not u:
                next_node = self._predecessors[u][next_node]
                seq.append(next_node)
            seq.append(u)
        return seq[::-1]

    @property
    def vertices(self):
        return list(self.G.nodes)

    @property
    def edges(self):
        return list(self.G.edges)

    @property
    def weights(self):
        return list(nx.get_edge_attributes(self.G, "distance").values())

    def random_point(self):
        return random.choice(self.vertices)

    def __repr__(self):
        return f"Graph(..., velocity={self.velocity})"

    def asdict(self):
        return dict(
            vertices=self.vertices,
            edges=self.edges,
            weights=self.weights,
            velocity=self.velocity,
        )

    def __reduce__(self):
        return self.__class__, (
            self.vertices,
            self.edges,
            self.weights,
            self.velocity,
        )


class DiGraph(Graph):
    def __init__(
        self,
        vertices: Sequence[int],
        edges: Sequence[Tuple[int, int]],
        weights: Union[None, float, Sequence[float]],
        velocity: float = 1,
    ):
        """
        Weighted directed graph with integer vertex labels

        Parameters
        ----------
        vertices
            sequence of vertices
        edges
            sequence of edge tuples
        weights
            Edge weights.
            - if None is supplied, the resulting graph is unweighted (unit edge length)
            - if a single float is supplied, every edge length will be equal to this number
            - if a sequence is supplied, this will be mapped onto the edge sequence
        velocity
            constant velocity to compute travel time, optional.
        """
        self.G = nx.DiGraph()
        self.G.add_nodes_from(vertices)
        self.G.add_edges_from(
            (u, v, {"distance": w}) for (u, v), w in zip(edges, weights)
        )

        self.velocity = velocity
        self._update_distance_cache()

    @classmethod
    def from_nx(
        cls,
        G: nx.Graph,
        velocity: float = 1,
        make_attribute_distance: str = "distance",
    ):
        """
        Create a graph space from networkx directed graph
        Parameters
        ----------
        G
            the networkx.DiGraph instance, will be deepcopied
        velocity
            velocity to use for travel time computation
        make_attribute_distance
            attribute to rename to "distance" and use as such
            If None is supplied, the resulting graph is unweighted (unit edge length).

        Returns
        -------
        graph space instance
        """
        self = cls.__new__(cls)
        self.G = cls._prepare_copy_of_nx_graph(
            G=G, make_attribute_distance=make_attribute_distance, directed=True
        )
        self.velocity = velocity
        self._update_distance_cache()
        return self

    def __repr__(self):
        return f"DiGraph(..., velocity={self.velocity})"


class ContinuousGraph(Graph):
    def __init__(self):
        raise NotImplementedError

    @smartVectorize
    def d(self, u, v):
        """coordinates shall consist of triples (u, v, dist_to_dest)"""
        ...

    def t(self, u, v): ...

    def interp_dist(self, u, w, time_to_dest): ...

    def interp_time(self, u, v, time_to_dest): ...

    def random_point(self): ...



#!/usr/bin/env python
import os

from collections import defaultdict
from Cython.Build import cythonize
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

BUILD_ARGS = defaultdict(lambda: ["-std=c++17"], msvc=["/std:c++17"])


class build_ext_w_compiler_check(build_ext):
    # Adapted from <https://stackoverflow.com/questions/30985862/how-to-identify- \
    #   compiler-before-defining-cython-extensions/32192172#32192172>
    def build_extensions(self):
        compiler = self.compiler.compiler_type
        args = BUILD_ARGS[compiler]
        for ext in self.extensions:
            ext.extra_compile_args = args
        build_ext.build_extensions(self)


setup(
    ext_modules=cythonize(
        [
            Extension(
                name="*",
                sources=["src/ridepy/**/*.pyx"],
                include_dirs=[
                    "src/lru-cache/include",
                    "src/ridepy/util/spaces_cython",
                    "src/ridepy/util/dispatchers_cython",
                    "src/ridepy/data_structures_cython",
                ],
            )
        ],
        compiler_directives={"embedsignature": True},
    ),
    options={
        "build_ext": {
            "inplace": True,
            "parallel": os.cpu_count() - 1,
        }
    },
    cmdclass={"build_ext": build_ext_w_compiler_check},
)



# distutils: language=c++

from .cdispatchers cimport AbstractDispatcher as CAbstractDispatcher
from ridepy.data_structures_cython.data_structures cimport LocType, R2loc


cdef union UDispatcher:
    CAbstractDispatcher[R2loc] *dispatcher_r2loc_ptr
    CAbstractDispatcher[int] *dispatcher_int_ptr

cdef class Dispatcher:
    cdef UDispatcher u_dispatcher
    cdef readonly LocType loc_type


import logging
import os
import warnings

import loky
from time import time

import operator as op
import functools as ft
import itertools as it
import pandas as pd

from collections import defaultdict
from copy import deepcopy
from typing import (
    Iterator,
    Any,
    Optional,
    Iterable,
    Union,
    Sequence,
    Callable,
    Mapping,
)
from pathlib import Path

from ridepy.util import make_sim_id
from ridepy.util.analytics import (
    get_system_quantities,
    get_vehicle_quantities,
    get_stops_and_requests_from_events_dataframe,
)
from ridepy.util.dispatchers import BruteForceTotalTravelTimeMinimizingDispatcher
from ridepy.util.dispatchers_cython import (
    BruteForceTotalTravelTimeMinimizingDispatcher as CyBruteForceTotalTravelTimeMinimizingDispatcher,
)
from ridepy.util.spaces import Euclidean2D
from ridepy.util.spaces_cython import Euclidean2D as CyEuclidean2D
from ridepy.util.request_generators import RandomRequestGenerator
from ridepy.data_structures import (
    TransportationRequest,
)
from ridepy.data_structures_cython import (
    TransportationRequest as CyTransportationRequest,
)
from ridepy.vehicle_state import VehicleState
from ridepy.vehicle_state_cython import VehicleState as CyVehicleState
from ridepy.fleet_state import SlowSimpleFleetState
from ridepy.extras.io import (
    create_params_json,
    save_events_json,
    read_params_json,
)

logger = logging.getLogger(__name__)


def make_file_path(sim_id: str, directory: Path, suffix: str) -> Path:
    return directory / f"{sim_id}{suffix}"


def get_params(
    directory: Path, sim_id: str, param_path_suffix: str = "_params.json"
) -> dict:
    return read_params_json(make_file_path(sim_id, directory, param_path_suffix))


def perform_single_analysis(
    sim_id: str,
    data_dir: Path,
    tasks_if_not_existent: Optional[Iterable[str]] = None,
    tasks_if_existent: Optional[Iterable[str]] = None,
    param_path_suffix: str = "_params.json",
    event_path_suffix: str = ".jsonl",
    stops_path_suffix: str = "_stops.pq",
    requests_path_suffix: str = "_requests.pq",
    vehicle_quantities_path_suffix: str = "_vehicle_quantities.pq",
) -> tuple[str, dict[str, Any]]:
    """
    Compute stops, requests, vehicle quantities, and system quantities from
    simulation events.

    Parameters
    ----------
    sim_id
        Simulation ID
    data_dir
        Directory from which to read the events and to which to write the resulting
        parquet output files
    tasks_if_not_existent
        Collection of tasks to perform if the respective output files do not exist.
        Valid entries are 'stops', 'requests', 'vehicle_quantities', and
        'system_quantities'. If None, everything is computed if not existent.
    tasks_if_existent
        Collection of tasks to perform even if the respective output files do exist,
        i.e. tasks to recompute. Valid entries are 'stops', 'requests',
        'vehicle_quantities', and 'system_quantities'. If None, nothing is recomputed.
    param_path_suffix
        Appending this suffix to the simulation ID yields the path to the JSON parameter
        file.
    event_path_suffix
        Appending this suffix to the simulation ID yields the path to the JSONL
        event file.
    stops_path_suffix
        Appending this suffix to the simulation ID yields the path to the parquet
        stops file.
    requests_path_suffix
        Appending this suffix to the simulation ID yields the path to the parquet
        requests file.
    vehicle_quantities_path_suffix
        Appending this suffix to the simulation ID yields the path to the \
        parquet vehicle quantities file.

    Returns
    -------
    sim_id
        Simulation ID
    system_quantities
        Dictionary containing the computed system quantities as entries
    """
    stops_path = make_file_path(sim_id, data_dir, stops_path_suffix)
    requests_path = make_file_path(sim_id, data_dir, requests_path_suffix)
    vehicle_quantities_path = make_file_path(
        sim_id, data_dir, vehicle_quantities_path_suffix
    )

    tasks = set()

    tasks_if_not_existent = tasks_if_not_existent or {
        "stops",
        "requests",
        "vehicle_quantities",
        "system_quantities",
    }
    tasks_if_existent = tasks_if_existent or set()

    if "stops" in tasks_if_not_existent and not stops_path.exists():
        tasks.add("stops")
    if "requests" in tasks_if_not_existent and not requests_path.exists():
        tasks.add("requests")
    if (
        "vehicle_quantities" in tasks_if_not_existent
        and not vehicle_quantities_path.exists()
    ):
        tasks.add("vehicle_quantities")
    if "system_quantities" in tasks_if_not_existent:
        tasks.add("system_quantities")

    tasks |= tasks_if_existent

    system_quantities = {}
    if tasks:
        params = get_params(data_dir, sim_id, param_path_suffix=param_path_suffix)
        space = params["general"]["space"]

        if "stops" in tasks or "requests" in tasks:
            stops, requests = get_stops_and_requests_from_events_dataframe(
                events_df=pd.read_json(
                    make_file_path(sim_id, data_dir, event_path_suffix), lines=True
                ),
                space=space,
            )
        else:
            stops = pd.read_parquet(stops_path)
            requests = pd.read_parquet(requests_path)

        if "vehicle_quantities" in tasks:
            vehicle_quantities = get_vehicle_quantities(stops, requests)
            vehicle_quantities.to_parquet(vehicle_quantities_path)

        if "system_quantities" in tasks:
            system_quantities = get_system_quantities(stops, requests, params)

        if "stops" in tasks:
            if space.n_dim > 1:
                stops["location"] = stops[~stops["location"].isna()]["location"].map(
                    list
                )
            stops.to_parquet(stops_path)

        if "requests" in tasks:
            if space.n_dim > 1:
                if ("submitted", "origin") in requests:
                    requests["submitted", "origin"] = requests[
                        ~requests["submitted", "origin"].isna()
                    ]["submitted", "origin"].map(list)

                if ("accepted", "origin") in requests:
                    requests["accepted", "origin"] = requests[
                        ~requests["accepted", "origin"].isna()
                    ]["accepted", "origin"].map(list)

                if ("submitted", "destination") in requests:
                    requests["submitted", "destination"] = requests[
                        ~requests["submitted", "destination"].isna()
                    ]["submitted", "destination"].map(list)

                if ("accepted", "destination") in requests:
                    requests["accepted", "destination"] = requests[
                        ~requests["accepted", "destination"].isna()
                    ]["accepted", "destination"].map(list)

            requests.to_parquet(requests_path)

    return sim_id, system_quantities


def perform_single_simulation(
    params: dict[str, dict[str, Any]],
    *,
    data_dir: Path,
    jsonl_chunksize: int = 1000,
    debug: bool = False,
    param_path_suffix: str = "_params.json",
    event_path_suffix: str = ".jsonl",
    dry_run: bool = False,
) -> str:
    """
    Execute a single simulation run based on a parameter dictionary
    and save parameters and result events to disk.

    Parameters
    ----------
    params
        Parameter dictionary to base the simulation on. Must contain the following keys:

        - ``general``
            - either ``n_reqs``  or ``t_cutoff``
            - ``seat_capacity``
            - (``initial_location`` and ``n_vehicles``) or ``initial_locations``
            - ``space``
            - ``transportation_request_cls``
            - ``vehicle_state_cls``
            - ``fleet_state_cls``
        - ``request_generator``
            - ``request_generator_cls``
        - ``dispatcher``
            - ``dispatcher_cls``
            - ...
    data_dir
        Existing directory in which to store parameters and events.
    jsonl_chunksize
        Number of simulation events to keep in memory before writing them to disk at once.
    debug
        Print debug info to stdout.
    param_path_suffix
        Parameters will be stored under "data_dir/<simulation_id><suffix>"
    event_path_suffix
        Simulation events will be stored under "data_dir/<simulation_id><event_path_suffix>"
    dry_run
        If True, do not actually simulate. Just pretend to and return the corresponding ID.

    Returns
    -------
    simulation ID
    """
    # we need a pseudorandom id that does not change if this function is called with the same params
    # the following does not guarantee a lack of collisions, and will fail if non-ascii characters are involved.
    tick = time()
    params_json = create_params_json(params=params)
    sim_id = make_sim_id(params_json)
    event_path = data_dir / f"{sim_id}{event_path_suffix}"
    param_path = data_dir / f"{sim_id}{param_path_suffix}"

    if (
        param_path.exists()
        and (params_json_ := param_path.read_text())
        and params_json_[-1] == "}"
    ):
        # assume that a previous simulation run already exists. this works because we write
        # to param_path *after* a successful simulation run.
        logger.info(
            f"Pre-existing param json exists for {params_json=} at {param_path=}, skipping simulation"
        )
        return sim_id
    else:
        logger.info(
            f"No pre-existing param json exists for {params_json=} at {param_path=}, running simulation"
        )
        if event_path.exists():
            logger.info(
                f"Potentially incomplete simulation data exists at {event_path=}, this will be overwritten"
            )
            event_path.unlink()

    space = params["general"]["space"]
    request_generator_cls = params["request_generator"].pop("request_generator_cls")
    rg = request_generator_cls(
        space=space,
        request_class=params["general"]["transportation_request_cls"],
        **params["request_generator"],
    )

    dispatcher = params["dispatcher"].pop("dispatcher_cls")
    if (
        params["general"].get("n_vehicles") is not None
        and params["general"].get("initial_location") is not None
        and params["general"].get("initial_locations") is None
    ):
        initial_locations = {
            vehicle_id: params["general"]["initial_location"]
            for vehicle_id in range(params["general"]["n_vehicles"])
        }
    elif (
        params["general"].get("n_vehicles") is None
        and params["general"].get("initial_location") is None
        and params["general"].get("initial_locations") is not None
    ):
        initial_locations = params["general"]["initial_locations"]
    else:
        raise ValueError(
            "Must *either* specify `n_vehicles` *and* `initial_location` *or* `initial_locations`"
        )

    fs = params["general"]["fleet_state_cls"](
        initial_locations=initial_locations,
        space=space,
        dispatcher=dispatcher(loc_type=space.loc_type, **params["dispatcher"]),
        seat_capacities=params["general"]["seat_capacity"],
        vehicle_state_class=params["general"]["vehicle_state_cls"],
    )

    # NOTE: this string is matched for testing
    if debug:
        print(f"Simulating run on process {os.getpid()} @ \n{params!r}\n")

    if not dry_run:
        if (
            params["general"]["n_reqs"] is not None
            and params["general"]["t_cutoff"] is None
        ):
            simulation = fs.simulate(it.islice(rg, params["general"]["n_reqs"]))
        elif (
            params["general"]["n_reqs"] is None
            and params["general"]["t_cutoff"] is not None
        ):
            simulation = fs.simulate(rg, t_cutoff=params["general"]["t_cutoff"])
        else:
            raise ValueError("Must *either* specify `n_reqs` *or* `t_cutoff`")

        while chunk := list(it.islice(simulation, jsonl_chunksize)):
            save_events_json(jsonl_path=event_path, events=chunk)

        with open(str(param_path), "w") as f:
            f.write(params_json)
    tock = time()
    if debug:
        print(f"Simulation run on process {os.getpid()} took {tock-tick} seconds\n")
    return sim_id


def simulate_parameter_combinations(
    *,
    param_combinations: Iterator[dict[str, dict[str, Sequence[Any]]]],
    data_dir: Union[str, Path],
    debug: bool = False,
    max_workers: Optional[int] = None,
    process_chunksize: int = 1,
    jsonl_chunksize: int = 1000,
    event_path_suffix: str = ".jsonl",
    param_path_suffix: str = "_params.json",
    dry_run: bool = False,
):
    """
    Run simulations for different parameter combinations using multiprocessing.

    Parameters
    ----------
    param_combinations
        An iterable of parameter configurations. For more detail see :ref:`Executing Simulations`
    data_dir
        Directory in which to store the parameters and events.
    debug
        Print debug info.
    max_workers
        Maximum number of multiprocessing workers. Defaults to number of processors
        on the machine if `None` or not given.
    process_chunksize
        Number of simulations to submit to each multiprocessing worker at a time.
    jsonl_chunksize
        Maximum number of events to keep in memory before saving to disk
    param_path_suffix
        Parameters will be stored under "data_dir/<simulation_id><suffix>"
    event_path_suffix
        Simulation events will be stored under "data_dir/<simulation_id><event_path_suffix>"
    dry_run
        If True, do not actually simulate. Just pretend to and return the corresponding IDs.

    Returns
    -------
        List of simulation IDs. See the docstring of `.SimulationSet` for more detail.
    """
    with loky.get_reusable_executor(max_workers=max_workers) as executor:
        sim_ids = list(
            executor.map(
                ft.partial(
                    perform_single_simulation,
                    debug=debug,
                    jsonl_chunksize=jsonl_chunksize,
                    data_dir=data_dir,
                    param_path_suffix=param_path_suffix,
                    event_path_suffix=event_path_suffix,
                    dry_run=dry_run,
                ),
                param_combinations,
                chunksize=process_chunksize,
            )
        )
    return sim_ids


class SimulationSet:
    """
    A set of simulations. The parameter space is defined through constant `base_params`,
    zipped `zip_params` and cartesian product `product_params`. A set of a single simulation
    is also allowed. Through `SimulationSet.run`, configurable multiprocessing is implemented,
    allowing for parallelization of simulation runs at different parameters.
    """

    @staticmethod
    def _two_level_dict_update(
        base_dict: dict[str, dict[str, Any]], update_dict: dict[str, dict[str, Any]]
    ) -> dict:
        """
        Update two-level nested dictionary with deepcopying,
        where `update_dict` overwrites entries in `base_dict`.

        Example
        -------

        .. code-block:: python

            SimulationSet._two_level_dict_update(
                {"a": {"b": 5, 6: 8}}, {"a": {6: "fooo", 8: "baaar"}, "baz": {"6": 6}}
            )

        yields

        .. code-block:: python

            {"a": {"b": 5, 6: "fooo", 8: "baaar"}, "baz": {"6": 6}}


        Parameters
        ----------
        base_dict
            two-level nested dict to update
        update_dict
            two-level nested dict which overwrites entries in base_dict

        Returns
        -------
        updated deepcopy of base_dict

        """
        d = deepcopy(base_dict)
        # This sorted is needed otherwise detection of pre existing simulation run does not work.
        for outer_key in sorted(set(base_dict) | set(update_dict)):
            d[outer_key] = base_dict.get(outer_key, {}) | update_dict.get(outer_key, {})
        return d

    @staticmethod
    def _zip_params_equal_length(
        zip_params: dict[str, dict[str, Sequence[Any]]]
    ) -> bool:
        """
        Evaluate whether the sequences in the inner dict are of equal length.

        Parameters
        ----------
        zip_params

        Returns
        -------
        True if sequences are of equal length, False otherwise

        """

        if zip_params:
            g = it.groupby(
                len(inner_value)
                for inner_dict in zip_params.values()
                for inner_value in inner_dict.values()
            )
            return next(g, True) and not next(g, False)
        else:
            return True

    @property
    def data_dir(self) -> Path:
        """
        Get directory in which to store the parameters and events.
        """
        return self._data_dir

    @data_dir.setter
    def data_dir(self, data_dir: Union[str, Path]) -> None:
        """
        Set directory in which to store the parameters and events.
        Will be created if not existent.
        """
        data_dir = Path(data_dir)
        data_dir.mkdir(exist_ok=True, parents=True)
        self._data_dir = data_dir

    def __init__(
        self,
        *,
        data_dir: Union[str, Path],
        base_params: Optional[dict[str, dict[str, Any]]] = None,
        zip_params: Optional[dict[str, dict[str, Sequence[Any]]]] = None,
        product_params: Optional[dict[str, dict[str, Sequence[Any]]]] = None,
        cython: bool = True,
        debug: bool = False,
        max_workers: Optional[int] = None,
        process_chunksize: int = 1,
        jsonl_chunksize: int = 1000,
        event_path_suffix: str = ".jsonl",
        param_path_suffix: str = "_params.json",
        validate: bool = True,
    ) -> None:
        """

        Parameters
        ----------
        data_dir
            Directory in which to store the parameters and events.
        base_params
            Dictionary setting parameters that are kept constant throughout the simulation set, optional.
        zip_params
            Dictionary setting parameters that are varied together throughout the simulation set, optional.
            The values for each inner dict should be lists that all match in lengths.
        product_params
            Dictionary setting parameters of which the cartesian product (i.e. all possible
            combinations of the supplied parameters) is created and varied throughout
            the simulation set, optional.
        cython
            Use cython.
        debug
            Print debug info.
        max_workers
            Maximum number of multiprocessing workers. Defaults to number of processors
            on the machine if `None` or not given.
        process_chunksize
            Number of simulations to submit to each multiprocessing worker at a time.
        jsonl_chunksize
            Maximum number of events to keep in memory before saving to disk
        param_path_suffix
            Parameters will be stored under "data_dir/<simulation_id><suffix>"
        event_path_suffix
            Simulation events will be stored under "data_dir/<simulation_id><event_path_suffix>"
        validate
            Check validity of the supplied dictionary (unknown outer and inner keys, equal length for ``zip_params``)
        """

        self.debug = debug
        self.max_workers = max_workers
        self.process_chunksize = process_chunksize
        self.jsonl_chunksize = jsonl_chunksize
        self.data_dir = Path(data_dir)

        self._event_path_suffix = event_path_suffix
        self._param_path_suffix = param_path_suffix

        if cython:
            space_obj = CyEuclidean2D()
            dispatcher = CyBruteForceTotalTravelTimeMinimizingDispatcher
            transportation_request_cls = CyTransportationRequest
            vehicle_state_cls = CyVehicleState
        else:
            space_obj = Euclidean2D()
            dispatcher = BruteForceTotalTravelTimeMinimizingDispatcher
            transportation_request_cls = TransportationRequest
            vehicle_state_cls = VehicleState

        self.default_base_params = dict(
            general=dict(
                n_reqs=100,
                t_cutoff=None,
                space=space_obj,
                n_vehicles=10,
                initial_location=(0, 0),
                initial_locations=None,
                seat_capacity=8,
                transportation_request_cls=transportation_request_cls,
                vehicle_state_cls=vehicle_state_cls,
                fleet_state_cls=SlowSimpleFleetState,
            ),
            dispatcher=dict(dispatcher_cls=dispatcher),
            request_generator=dict(
                request_generator_cls=RandomRequestGenerator,
                rate=1,
            ),
        )

        base_params = base_params if base_params is not None else {}
        zip_params = zip_params if zip_params is not None else {}
        product_params = product_params if product_params is not None else {}

        if validate:
            # assert no unknown outer keys
            assert not (set(base_params) | set(zip_params) | set(product_params)) - set(
                self.default_base_params
            ), "invalid outer key"

            # assert no unknown inner keys
            for outer_key in set(self.default_base_params) - {
                "request_generator",
                "dispatcher",
            }:
                assert not (
                    set(base_params.get(outer_key, {}))
                    | set(zip_params.get(outer_key, {}))
                    | set(product_params.get(outer_key, {}))
                ) - set(
                    self.default_base_params[outer_key]
                ), f"invalid inner key for {outer_key=}"

            # assert equal length of zipped parameters
            assert self._zip_params_equal_length(
                zip_params
            ), "zipped parameters must be of equal length"

        self._base_params = self._two_level_dict_update(
            self.default_base_params, base_params
        )
        self._zip_params = zip_params
        self._product_params = product_params

        self._simulation_ids = None

        self._system_quantities_path = None

    @property
    def simulation_ids(self) -> list[str]:
        """
        Get simulation IDs.
        """
        # protect simulation ids
        return self._simulation_ids if self._simulation_ids is not None else []

    @property
    def param_paths(self) -> list[Path]:
        """
        Get list of JSON parameter files.
        """
        return [
            self.data_dir / f"{simulation_id}{self._param_path_suffix}"
            for simulation_id in self.simulation_ids
        ]

    @property
    def event_paths(self) -> list[Path]:
        """
        Get list of resulting output event JSON Lines file paths.
        """
        return [
            self.data_dir / f"{simulation_id}{self._event_path_suffix}"
            for simulation_id in self.simulation_ids
        ]

    @staticmethod
    def _make_joined_key_pairs_values(
        *, params: dict[str, dict[str, Sequence[Any]]], join_fn: Callable
    ) -> tuple[Iterable[tuple[str, str]], Iterable[Iterable[Any]]]:
        """

        Parameters
        ----------
        params
            Parameter dictionary containing sequences as inner values.
        join_fn
            e.g. `zip` or `itertools.product`

        Returns
        -------
        joined_key_pairs, joined_values_iter
        """
        if params:
            joined_values_iter = join_fn(
                *(
                    params[outer_key][inner_key]
                    for outer_key in params.keys()
                    for inner_key in params[outer_key].keys()
                )
            )
            joined_key_pairs = [
                (outer_key, inner_key)
                for outer_key in params.keys()
                for inner_key in params[outer_key].keys()
            ]
        else:
            joined_values_iter = joined_key_pairs = [tuple()]

        return joined_key_pairs, joined_values_iter

    def __iter__(self):
        zipped_key_pairs, zipped_values_iter = self._make_joined_key_pairs_values(
            params=self._zip_params, join_fn=zip
        )

        (
            multiplied_key_pairs,
            multiplied_values_iter,
        ) = self._make_joined_key_pairs_values(
            params=self._product_params, join_fn=it.product
        )

        def param_combinations() -> Iterator[dict[str, dict[str, Any]]]:
            """
            Generator yielding complete parameter sets which can be
            supplied to `perform_single_simulation`.
            """
            for zipped_params, multiplied_params in it.product(
                zipped_values_iter, multiplied_values_iter
            ):
                d = defaultdict(dict)

                for (outer_key, inner_key), value in zip(
                    zipped_key_pairs, zipped_params
                ):
                    d[outer_key][inner_key] = value

                for (outer_key, inner_key), value in zip(
                    multiplied_key_pairs, multiplied_params
                ):
                    d[outer_key][inner_key] = value

                yield self._two_level_dict_update(self._base_params, d)

        self._param_combinations = param_combinations()
        return self

    def __next__(self):
        return next(self._param_combinations)

    def run(self, dry_run=False):
        """
        Run the simulations configured through `base_params`, `zip_params` and `product_params` using
        multiprocessing. The parameters and resulting output events are written to disk
        in JSON/JSON Lines format. For more detail see :ref:`Executing Simulations`.

        Access simulations results
            - by id: `SimulationSet.simulation_ids`
            - by parameter file `SimulationSet.param_paths`
            - by event file `SimulationSet.event_paths`

        Parameters
        ----------
        dry_run
            If True, do not actually simulate.
        """

        self._simulation_ids = simulate_parameter_combinations(
            param_combinations=iter(self),
            data_dir=self.data_dir,
            debug=self.debug,
            max_workers=self.max_workers,
            process_chunksize=self.process_chunksize,
            jsonl_chunksize=self.jsonl_chunksize,
            event_path_suffix=self._event_path_suffix,
            param_path_suffix=self._param_path_suffix,
            dry_run=dry_run,
        )

    def __len__(self) -> int:
        """
        Number of simulations performed when calling `SimulationSet.run`.
        """
        len_ = 1
        if self._zip_params:
            len_ *= len(next(iter(next(iter(self._zip_params.values())).values())))
        if self._product_params:
            len_ *= ft.reduce(
                op.mul,
                (
                    len(inner_value)
                    for inner_dict in self._product_params.values()
                    for inner_value in inner_dict.values()
                ),
            )
        if not (self._zip_params or self._product_params):
            len_ = 0

        return len_

    def run_analytics(
        self,
        only_stops_and_requests: bool = False,
        update_existing: Union[bool, list[str]] = False,
        check_for_changes: bool = True,
        stops_path_suffix: str = "_stops.pq",
        requests_path_suffix: str = "_requests.pq",
        vehicle_quantities_path_suffix: str = "_vehicle_quantities.pq",
        system_quantities_filename: str = "system_quantities.pq",
    ) -> None:
        """
        Compute analytics from simulation events and store them to disk
        in parquet format.

        Parameters
        ----------
        only_stops_and_requests
            Only compute stops and requests, not vehicle and system quantities.
        update_existing
            Recompute existing outputs. If a list is given, only recompute
            the list entries. Valid list items are 'system_quantities',
            'vehicle_quantities', 'stops', and 'requests'.
        check_for_changes
            If True, only update system quantities if simulation ids have changed.
            If False, do update system quantities in any case.
        stops_path_suffix
            Appending this suffix to the simulation ID yields the path to the parquet
            stops file.
        requests_path_suffix
            Appending this suffix to the simulation ID yields the path to the parquet
            requests file.
        vehicle_quantities_path_suffix
            Appending this suffix to the simulation ID yields the path to the parquet
            vehicle quantities file.
        system_quantities_filename
            Filename of the parquet file to store the system quantities in.
        """
        self._system_quantities_path = self.data_dir / system_quantities_filename

        if not self.simulation_ids:
            warnings.warn(
                "no simulations have been run (simulation_ids empty)", UserWarning
            )
        else:

            # In any case, stops and requests have to be computed if they don't exist
            tasks_if_not_existent = {"stops", "requests"}

            if not only_stops_and_requests:
                # Additionally, we are computing vehicle and system quantities, now,
                # if they don't exist
                tasks_if_not_existent |= {"vehicle_quantities", "system_quantities"}

            if isinstance(update_existing, Iterable):
                # If we have been handed a list of tasks to update, we only update these,
                # not recomputing tasks that should not have be computed in the first place
                tasks_if_existent = set(update_existing) & tasks_if_not_existent
            elif update_existing == True:
                tasks_if_existent = tasks_if_not_existent
            elif update_existing == False:
                tasks_if_existent = set()
            else:
                raise ValueError(f"Got invalid value for {update_existing=}")

            if self._system_quantities_path.exists():
                tasks_if_not_existent -= {"system_quantities"}
                if check_for_changes:
                    # Currently, we only check for changes in the sense that new
                    # simulations with new ids have been performed and are thus missing from
                    # the system quantities output. For vehicle quantities, stops, and requests,
                    # check_for_changes does not apply, as we compute these in any case, should
                    # they be missing.
                    sqdf = pd.read_parquet(self._system_quantities_path)
                    if set(sqdf.index) == set(self.simulation_ids):
                        tasks_if_existent -= {"system_quantities"}
                    del sqdf

            with loky.get_reusable_executor(max_workers=self.max_workers) as executor:
                sim_ids, system_quantities = zip(
                    *list(
                        executor.map(
                            ft.partial(
                                perform_single_analysis,
                                data_dir=self.data_dir,
                                tasks_if_not_existent=tasks_if_not_existent,
                                tasks_if_existent=tasks_if_existent,
                                param_path_suffix=self._param_path_suffix,
                                event_path_suffix=self._event_path_suffix,
                                stops_path_suffix=stops_path_suffix,
                                requests_path_suffix=requests_path_suffix,
                                vehicle_quantities_path_suffix=vehicle_quantities_path_suffix,
                            ),
                            self.simulation_ids,
                            chunksize=self.process_chunksize,
                        )
                    )
                )

            if list(sim_ids) != self.simulation_ids:
                warnings.warn(
                    "Simulation IDs and IDs analytics were computed for do not match. Something's not right.",
                    UserWarning,
                )

            if "system_quantities" in tasks_if_not_existent | tasks_if_existent:
                system_quantities_df = pd.DataFrame(system_quantities, index=sim_ids)
                system_quantities_df.rename_axis("simulation_id", inplace=True)
                system_quantities_df.to_parquet(self._system_quantities_path)

    @property
    def param_path_suffix(self) -> str:
        """
        Get the parameter file suffix.
        """
        return self._param_path_suffix

    @property
    def event_path_suffix(self) -> str:
        """
        Get the event file suffix.
        """
        return self._event_path_suffix

    @property
    def system_quantities_path(self) -> Path:
        """
        Get path to the parquet file containing the system quantities.

        Returns
        -------
        system_quantities_path
        """
        if self._system_quantities_path is not None:
            return self._system_quantities_path
        else:
            raise AttributeError("No system quantities path set.")

    def get_system_quantities(
        self, extra_params: Optional[Mapping] = None
    ) -> pd.DataFrame:
        """
        Return the system quantities, if computed already using `run_analytics`.

        Optionally, join the system quantities dataframe with additional arbitrary
        parameters used in the simulation.

        Parameters
        ----------
        extra_params:
            Dictionary specifying the desired parameters. The keys are the resulting column
            names in the returned dataframe, the values are the (nested) keys in the
            `SimulationSet` params dictionary, joined by ``.``.

            Example:

            .. code-block:: python

                extra_params = {"n": "general.n_reqs"}

        Returns
        -------
        system_quantities
            DataFrame containing the system quantities and optional extra parameters

        """

        try:
            sqdf = pd.read_parquet(self.system_quantities_path)
        except FileNotFoundError:
            raise FileNotFoundError(
                f"System properties file {self.system_quantities_path} not found."
            )

        if extra_params:
            for sim_id in sqdf.index:
                params = get_params(
                    self.data_dir, sim_id, param_path_suffix=self.param_path_suffix
                )
                for param_col, param_key in extra_params.items():
                    sqdf.loc[sim_id, param_col] = ft.reduce(
                        op.getitem, param_key.split("."), params
                    )

        return sqdf



import dataclasses

from ridepy.data_structures import Location, TransportSpace, Stoplist, Dispatcher
from ridepy.util.spaces_cython import TransportSpace as CyTransportSpace
from ridepy.util.testing_utils_cython import (
    BruteForceTotalTravelTimeMinimizingDispatcher as CyBruteForceTotalTravelTimeMinimizingDispatcher,
)
from ridepy.util.spaces_cython import spaces as cyspaces
from typing import Literal, Iterable, Union, Callable, Sequence

from ridepy import data_structures as pyds, data_structures_cython as cyds
from ridepy import data_structures_cython as cyds
from ridepy.util import spaces as pyspaces
from ridepy.util.dispatchers.ridepooling import (
    BruteForceTotalTravelTimeMinimizingDispatcher,
)


def stoplist_from_properties(
    *,
    stoplist_properties: Iterable[Sequence[Union[Location, float]]],
    space: Union[TransportSpace, CyTransportSpace],
    kind: str,
) -> Union[Stoplist, cyds.Stoplist]:
    """
    Generate stoplist from an iterable of stop properties.
    Format:

    .. code-block:: python

        (
            (location, CPAT, timewindow_min, timewindow_max),
            ...,
        )

    or

    .. code-block:: python

        (
            (location, CPAT, timewindow_min, timewindow_max, occupancy_after_servicing),
            ...,
        )

    Parameters
    ----------
    stoplist_properties
        Iterable of stop property tuples `(location, CPAT, timewindow_min, timewindow_max)`
    space
        Space to place the stops on
    kind
        "cython" or "python"

    Returns
    -------

    """

    if kind == "python":
        data_structure_module = pyds
    elif kind == "cython":
        data_structure_module = cyds
    else:
        raise ValueError(f"Supplied invalid {kind=}, must be 'python' or 'cython'")

    python_stoplist = [
        data_structure_module.Stop(
            location=prop[0],
            request=data_structure_module.InternalRequest(
                request_id=-1, creation_timestamp=0, location=prop[0]
            ),
            action=data_structure_module.StopAction.internal,
            estimated_arrival_time=prop[1],
            occupancy_after_servicing=prop[4] if len(prop) == 5 else 0,
            time_window_min=prop[2],
            time_window_max=prop[3],
        )
        for prop in stoplist_properties
    ]

    if kind == "python":
        return python_stoplist
    else:
        return cyds.Stoplist(python_stoplist=python_stoplist, loc_type=space.loc_type)


def setup_insertion_data_structures(
    *,
    stoplist_properties: Iterable[Sequence[Union[Location, float]]],
    request_properties,
    space_type: str,
    kind: str,
) -> tuple[
    Union[TransportSpace, CyTransportSpace],
    Union[pyds.TransportationRequest, cyds.TransportationRequest],
    Union[Stoplist, cyds.Stoplist],
    Dispatcher,
]:
    """

    Parameters
    ----------
    stoplist_properties
    request_properties
    space_type
    kind
        'cython' or 'python'

    Returns
    -------
    space, request, stoplist, dispatcher
    """

    if kind == "python":
        spaces = pyspaces
        ds = pyds
        dispatcher = BruteForceTotalTravelTimeMinimizingDispatcher
    elif kind == "cython":
        spaces = cyspaces
        ds = cyds
        dispatcher = CyBruteForceTotalTravelTimeMinimizingDispatcher
    else:
        raise ValueError(f"Supplied invalid {kind=}, must be 'python' or 'cython'")

    space = getattr(spaces, space_type)()

    # set up the request
    request = ds.TransportationRequest(**request_properties)

    # set up the stoplist
    stoplist = stoplist_from_properties(
        stoplist_properties=stoplist_properties, space=space, kind=kind
    )

    return space, request, stoplist, dispatcher(loc_type=space.loc_type)



# distutils: language=c++

from ridepy.data_structures_cython.data_structures cimport  LocType, R2loc

from .cdispatchers cimport (
AbstractDispatcher as CAbstractDispatcher,
BruteForceTotalTravelTimeMinimizingDispatcher as CBruteForceTotalTravelTimeMinimizingDispatcher,
SimpleEllipseDispatcher as CSimpleEllipseDispatcher,
)


cdef class Dispatcher:
    def __init__(self, loc_type, *args, **kwargs):
        self.loc_type = loc_type

    def __dealloc__(self):
        if self.loc_type == LocType.R2LOC:
            del self.u_dispatcher.dispatcher_r2loc_ptr
        elif self.loc_type == LocType.INT:
            del self.u_dispatcher.dispatcher_int_ptr
        else:
            raise ValueError("This line should never have been reached")

    def __reduce__(self):
        return self.__class__, (self.loc_type, )


cdef class BruteForceTotalTravelTimeMinimizingDispatcher(Dispatcher):
    def __cinit__(self, loc_type):
        if loc_type == LocType.R2LOC:
            self.u_dispatcher.dispatcher_r2loc_ptr = (
                new CBruteForceTotalTravelTimeMinimizingDispatcher[R2loc]()
                )
        elif loc_type == LocType.INT:
            self.u_dispatcher.dispatcher_int_ptr = (
                new CBruteForceTotalTravelTimeMinimizingDispatcher[int]()
            )
        else:
            raise ValueError("This line should never have been reached")


cdef class SimpleEllipseDispatcher(Dispatcher):
    def __cinit__(self, loc_type, max_relative_detour=0):
        if loc_type == LocType.R2LOC:
            self.u_dispatcher.dispatcher_r2loc_ptr = new CSimpleEllipseDispatcher[R2loc](max_relative_detour)
        elif loc_type == LocType.INT:
            self.u_dispatcher.dispatcher_int_ptr = new CSimpleEllipseDispatcher[int](max_relative_detour)
        else:
            raise ValueError("This line should never have been reached")


# distutils: language = c++

from libcpp.vector cimport vector
from libcpp.pair cimport pair

from .cspaces cimport (
    Euclidean2D as CEuclidean2D,
    TransportSpace as CTransportSpace,
    Manhattan2D as CManhattan2D,
    GraphSpace as CGraphSpace
)

from ridepy.data_structures_cython.data_structures cimport LocType, R2loc


cdef union USpace:
    CTransportSpace[R2loc] *space_r2loc_ptr
    CTransportSpace[int] *space_int_ptr


cdef class TransportSpace:
    cdef USpace u_space
    cdef readonly LocType loc_type
    cdef readonly int n_dim


cdef class Euclidean2D(TransportSpace):
    cdef CEuclidean2D *derived_ptr
    cdef readonly vector[pair[float, float]] coord_range


cdef class Manhattan2D(TransportSpace):
    cdef CManhattan2D *derived_ptr
    cdef readonly vector[pair[float, float]] coord_range

cdef class Grid2D(Manhattan2D):
    cdef readonly float n
    cdef readonly float m
    cdef readonly float dn
    cdef readonly float dm
    cdef readonly float velocity

# cdef class Grid2D_QM(Grid2D):
#     ...

cdef class Graph(TransportSpace):
    cdef CGraphSpace[int] *derived_ptr



# distutils: language=c++
# distutils: include_dirs = boost_1_75_0

from libcpp.pair cimport pair
from libcpp.vector cimport vector


cdef extern from "cspaces.cxx":
    pass

"""
As explained in further detail in https://stackoverflow.com/a/28727488,
virtual functions that are later overridden in a child class **must be declared only for the base class**.
That is, we must *not* declare Euclidean2D.d/t/... here, only in TransportSpace.
Doing so will result in an *ambiguous overridden function* error from cython.
"""

cdef extern from "ctransport_space.h" namespace 'ridepy':

    ctypedef pair[double, double] R2loc
    cdef cppclass TransportSpace[Loc]:
        TransportSpace()
        TransportSpace(double)
        double d(Loc u, Loc v)
        double t(Loc u, Loc v)
        pair[Loc, double] interp_dist(Loc u, Loc v, double dist_to_dest)
        pair[Loc, double] interp_time(Loc u, Loc v, double time_to_dest)


cdef extern from "cspaces.h" namespace 'ridepy':

    cdef cppclass Euclidean2D(TransportSpace[R2loc]):
        double velocity
        Euclidean2D()
        Euclidean2D(double)

    cdef cppclass Manhattan2D(TransportSpace[R2loc]):
        double velocity
        Manhattan2D()
        Manhattan2D(double)


cdef extern from "boost_graph_space.h" namespace 'ridepy':
    cdef cppclass GraphSpace[Loc](TransportSpace[Loc]):
        double velocity
        ctypedef pair[Loc, Loc] Edge

        GraphSpace(double, vector[Loc] vertex_vec, vector[Edge] edge_vec, vector[double] weight_vec)
        GraphSpace(double, vector[Loc] vertex_vec, vector[Edge] edge_vec)

        vector[Loc] get_vertices()
        vector[pair[Loc, Loc]] get_edges()
        vector[double] get_weights()